@@ -6,6 +6,7 @@ Many docstrings in this file are based on the PEP, which is in the public domain
 """
 
 import base64
+from collections import namedtuple, OrderedDict
 import datetime
 import math
 import re
@@ -31,8 +32,6 @@ _logger = logging.getLogger(__name__)
 
 _TIMESTAMP_PATTERN = re.compile(r'(\d+-\d+-\d+ \d+:\d+:\d+(\.\d{,6})?)')
 
-_ERROR_MSG_HEADER = "X-Thriftserver-Error-Message"
-
 ssl_cert_parameter_map = {
     "none": CERT_NONE,
     "optional": CERT_OPTIONAL,
@@ -49,6 +48,12 @@ _retry_policy = {                          # (type, default, min, max)
     "_retry_stop_after_attempts_duration":  (float, 900, 1, 86400),
 }
 
+
+THRIFT_ERROR_MESSAGE_HEADER = "x-thriftserver-error-message"
+DATABRICKS_ERROR_OR_REDIRECT_HEADER = "x-databricks-error-or-redirect-message"
+DATABRICKS_REASON_HEADER = "x-databricks-reason-phrase"
+
+
 def _parse_timestamp(value):
     if value:
         match = _TIMESTAMP_PATTERN.match(value)
@@ -246,8 +251,8 @@ class Connection(object):
                 USER_AGENT_NAME, __version__, kwargs.get("_user_agent_entry"))
 
         self._transport.setCustomHeaders({
-            "Authorization" : authorization_header,
-            "User-Agent" : useragent_header
+            "Authorization": authorization_header,
+            "User-Agent": useragent_header
         })
         protocol = thrift.protocol.TBinaryProtocol.TBinaryProtocol(self._transport)
         self._client = TCLIService.Client(protocol)
@@ -296,6 +301,48 @@ class Connection(object):
     def rollback(self):
         raise NotSupportedError("Databricks does not have transactions")
 
+    @staticmethod
+    def _extract_error_message_from_headers(headers):
+        err_msg = None
+        if THRIFT_ERROR_MESSAGE_HEADER in headers:
+            err_msg = headers[THRIFT_ERROR_MESSAGE_HEADER]
+        if DATABRICKS_ERROR_OR_REDIRECT_HEADER in headers:
+            if err_msg:  # We don't expect both to be set, but log both here just in case
+                err_msg = "Thriftserver message: {}, Databricks message: {}".format(err_msg, headers[DATABRICKS_ERROR_OR_REDIRECT_HEADER])
+            else:
+                err_msg = headers[DATABRICKS_ERROR_OR_REDIRECT_HEADER]
+            if DATABRICKS_REASON_HEADER in headers:
+                err_msg += ": " + headers[DATABRICKS_REASON_HEADER]
+        return err_msg
+
+    def _handle_request_error(self, error_info, attempt, elapsed):
+        max_attempts = self._retry_stop_after_attempts_count
+        max_duration_s = self._retry_stop_after_attempts_duration
+
+        if error_info.retry_delay is not None and elapsed + error_info.retry_delay > max_duration_s:
+            no_retry_reason = common.NoRetryReason.OUT_OF_TIME
+        elif error_info.retry_delay is not None and attempt >= max_attempts:
+            no_retry_reason = common.NoRetryReason.OUT_OF_ATTEMPTS
+        elif error_info.retry_delay is None:
+            no_retry_reason = common.NoRetryReason.NOT_RETRYABLE
+        else:
+            no_retry_reason = None
+
+        full_error_info_str = error_info.full_info_logging_str(no_retry_reason, attempt,
+                                                               max_attempts, elapsed,
+                                                               max_duration_s)
+
+        if no_retry_reason is not None:
+            user_friendly_error_message = error_info.user_friendly_error_message(
+                no_retry_reason, attempt, elapsed)
+            _logger.info("{}: {}".format(user_friendly_error_message, full_error_info_str))
+
+            raise OperationalError(user_friendly_error_message, error_info.error)
+
+        _logger.info("Retrying request after error in {} seconds: {}".format(
+            error_info.retry_delay, full_error_info_str))
+        time.sleep(error_info.retry_delay)
+
     def _make_request(self, method, request):
         """Execute given request, attempting retries when receiving HTTP 429/503.
 
@@ -310,13 +357,13 @@ class Connection(object):
         # retries until success or final failure achieved.
 
         t0 = time.time()
+
         def get_elapsed():
             return time.time() - t0
 
         def extract_retry_delay(attempt):
             # encapsulate retry checks, returns None || delay-in-secs
             # Retry IFF 429/503 code + Retry-After header set
-            delay = None
             http_code = getattr(self._transport, "code", None)
             retry_after = getattr(self._transport, "headers", {}).get("Retry-After")
             if http_code in [429, 503] and retry_after:
@@ -329,16 +376,20 @@ class Connection(object):
 
         def attempt_request(attempt):
             # splits out lockable attempt, from delay & retry loop
-            # returns tuple: (method_return, delay_fn(), error, error_message)
-            # - non-None method_return -> success, return and be done
-            # - non-None retry_delay -> sleep delay before retry
-            # - error, error_message always set when available
             try:
-                return method(request), None, None, None
+                _logger.debug("Sending request: {}".format(request))
+                response = method(request)
+                _logger.debug("Received response: {}".format(response))
+                return response
             except Exception as error:
                 retry_delay = extract_retry_delay(attempt)
-                error_message = getattr(self._transport, "headers", {}).get(_ERROR_MSG_HEADER)
-                return None, retry_delay, error, error_message
+                error_message = Connection._extract_error_message_from_headers(getattr(self._transport, "headers", {}))
+                return common.RequestErrorInfo(error=error,
+                                               error_message=error_message,
+                                               retry_delay=retry_delay,
+                                               http_code=getattr(self._transport, "code", None),
+                                               method=method.__name__,
+                                               request=request)
 
         # The real work:
         # - for each available attempt:
@@ -347,7 +398,7 @@ class Connection(object):
         #       if available: bounded delay and retry
         #       if not: raise error
         max_attempts = self._retry_stop_after_attempts_count
-        max_elapsed = self._retry_stop_after_attempts_duration
+        max_duration_s = self._retry_stop_after_attempts_duration
 
         # use index-1 counting for logging/human consistency
         for attempt in range(1, max_attempts + 1):
@@ -355,28 +406,17 @@ class Connection(object):
             # We do not want threads to be simultaneously sharing the Thrift Transport
             # because we use its state to determine retries
             with self._request_lock:
-                response, retry_delay, error, error_message = attempt_request(attempt)
+                response_or_error_info = attempt_request(attempt)
             elapsed = get_elapsed()
 
             # conditions: success, non-retry-able, no-attempts-left, no-time-left, delay+retry
-            if response: # log nothing here, presume that main request logging covers
-                return response
+            if not isinstance(response_or_error_info, common.RequestErrorInfo):
+                # log nothing here, presume that main request logging covers
+                return response_or_error_info
 
-            log_base = "{}; Request: {} Error: {} Bounded-Retry-Delay: {} " \
-                    "Attempt: {}/{} Elapsed-Seconds: {}/{}".format(
-                    method.__name__, request, error_message, retry_delay, attempt, max_attempts,
-                    elapsed, max_elapsed)
-
-            if (retry_delay is None
-                    or attempt >= max_attempts
-                    or elapsed + retry_delay > self._retry_stop_after_attempts_duration):
-                _logger.error("Error during " + log_base)
-                raise OperationalError("Error during Thrift request" +
-                    (": " + error_message) if error_message else "",
-                    error)
-
-            _logger.info("Delaying {} seconds before retry after ".format(retry_delay) + log_base)
-            time.sleep(retry_delay)
+            error_info = response_or_error_info
+            # The error handler will either sleep or throw an exception
+            self._handle_request_error(error_info, attempt, elapsed)
 
 
 class Cursor(common.DBAPICursor):
@@ -545,7 +585,7 @@ class Cursor(common.DBAPICursor):
                      ttypes.TOperationState.UKNOWN_STATE,
                      ttypes.TOperationState.CANCELED_STATE,
                      ttypes.TOperationState.TIMEDOUT_STATE]:
-            raise OperationalError("Query execution failed.\nState: {}; Error code: {}; SQLSTATE: {}\nError message: {}\n"
+            raise OperationalError("Query execution failed. State: {}; Error code: {}; SQLSTATE: {}; Error message: {}"
                                    .format(ttypes.TOperationState._VALUES_TO_NAMES[state],
                                            resp.errorCode,
                                            resp.sqlState,
